# a model for GTC cooling status
from twisted.logger import Logger
from astropy.time import Time
import requests


class GTCCooler(object):
    extra_rpc_calls = ()
    log = Logger()
    settable_attributes = ()

    def __init__(self, emulate=True):
        self.telemetry_delay = 10
        self.emulate = emulate
        # no state machine needed
        self.machines = {}
        if self.emulate:
            self.temp_function = self.fake_temps
        else:
            self.temp_function = self.get_temps

    def get_temps(self):
        url = "http://161.72.18.76:1026/v2/entities/TemperatureSensor3"
        try:
            response = requests.get(
                url, headers={"fiware-service": "gtc", "fiware-servicePath": "/"}
            )
            response.raise_for_status()
            data = response.json()
            self.temps = [
                float(data["temperature1"]["value"]),
                float(data["temperature2"]["value"]),
                float(data["temperature3"]["value"]),
            ]
        except Exception as e:
            self.log.warn("Failed to get flow temperatures: {error}", error=e)
            self.temps = [None, None, None]
        return self.temps

    def fake_temps(self):
        return [12.3, 12.3, 5.0]

    def telemetry(self):
        ts = Time.now()
        ts.format = "iso"
        tel = dict(timestamp=ts)
        t1, t2, t3 = self.temp_function()
        tel["ccd_input"] = t3
        tel["cab_pre_filter"] = t2
        tel["cab_post_filter"] = t1
        tel["state"] = {}
        return tel



import os
import time
import six

from astropy.io import ascii
from astropy.io import fits
from astropy.time import Time
from twisted.logger import Logger

from ..utils.obsmodes import Idle, get_obsmode
from ..devices.ngc import NGCDevice, NGCEmulator
from ..machines import create_machine


class NGC(object):
    """
    Model for an ESO NGC controller
    """

    log = Logger()

    settable_attributes = ()
    extra_rpc_calls = (
        "start",
        "stop",
        "abort",
        "pon",
        "poff",
        "seq_start",
        "seq_stop",
        "trigger",
        "reset",
        "summary",
        "status",
        "setup",
        "online",
        "standby",
        "offline",
        "exit",
        "start_ngc_server",
        "stop_ngc_server",
        "setup_ngc_server",
        "load_setup",
        "add_hdu",
        "cmd",
    )

    def __init__(self, emulate=True):
        if emulate:
            self._dev = NGCEmulator()
        else:
            self._dev = NGCDevice()

        # NGC already has it's own state machine, so don't duplicate here
        self.machines = {}

    def online(self):
        """
        Moves NGC server into ONLINE state.

        When moving to this state the h/w is connected and the controller reset.
        """
        return self.cmd("online")

    def exit(self):
        """
        Shuts down NGC server.

        Once in this state it is impossible to talk to NGC server and it must be started
        using `NGC.start_ngc_server`
        """
        return self.cmd("exit")

    def offline(self):
        """
        Turns NGC server off, moving it into LOADED state.

        In this state, the server is running and you can communicate with it, but all sub-processes
        (e.g acquisition task) are shut down and hardware is disconnected.
        """
        return self.cmd("off")

    def standby(self):
        """
        Move NGC server to STANDBY state.

        In this state, the server is not connected to the hardware, but the sub-processes
        (e.g acquisition task) are enabled.
        """
        return self.cmd("standby")

    def stop_ngc_server(self):
        """
        Shuts down ESO NGC Server
        """
        try:
            output = self._dev.stop_ngc_sw()
        except Exception as e:
            self.log.error("could not start NGC software: " + str(e))
            return False

        for key in output:
            msg = output[key]
            self.log.info("cmd {}: {}".format(key, msg))
        return True

    def start_ngc_server(self, gui=False):
        """
        Run commands necessary to bring up the ESO NGC server
        """
        try:
            # wamp calls supply args as str, so force type
            gui = bool(gui)
            output = self._dev.start_ngc_sw(gui=gui)
        except Exception as e:
            self.log.error("could not start NGC software: " + str(e))
            return False

        for key in output:
            msg = output[key]
            self.log.info("cmd {}: {}".format(key, msg))
        return True

    def setup_ngc_server(self):
        """
        Load HiPERCAM specific settings
        """
        try:
            output = self._dev.setup_ngc_hipercam()
        except Exception as e:
            self.log.error("could not setup NGC for hipercam: " + str(e))
            return False

        for key in output:
            msg = output[key]
            self.log.info("cmd {}: {}".format(key, msg))
        return True

    def cmd(self, cmd, *args):
        self.log.info("cmd {}: {}".format(cmd, " ".join(args)))
        try:
            msg, cmd_ok = self._dev.send_command(cmd, *args)
            if not cmd_ok:
                self.log.error("cmd {} failed with msg {}".format(cmd, msg))
        except Exception as e:
            msg = "could not run command {}: {}".format(cmd, e)
            self.log.error("could not run command {}: {}".format(cmd, e))
            cmd_ok = False

        return msg, cmd_ok

    def telemetry(self):
        """
        Called periodically to provide status etc for the NGC
        """
        ts = Time.now()
        tel = dict(timestamp=ts)
        tel.update(self._dev.database_summary)
        frame_msg, ok = self._dev.send_command("status", "DET.FRAM2.NO")
        try:
            frame_no = int(frame_msg.split()[1])
        except:
            frame_no = 0
        if not ok:
            tel["exposure.frame"] = 0
        else:
            tel["exposure.frame"] = frame_no
        return tel

    def start(self, *args):
        return self.cmd("start")

    def stop(self, *args):
        return self.cmd("stop")

    def abort(self, *args):
        return self.cmd("abort")

    def pon(self, *args):
        return self.cmd("cldc 0 enable")

    def poff(self, *args):
        return self.cmd("cldc 0 disable")

    def seq_start(self, *args):
        return self.cmd("seq 0 start")

    def seq_stop(self, *args):
        return self.cmd("seq 0 stop")

    def trigger(self, *args):
        return self.cmd("seq trigger")

    def reset(self, *args):
        return self.cmd("reset")

    def summary(self, *args):
        return self._dev.database_summary

    def status(self, param_name=None):
        if param_name is None:
            return self.cmd("status")

        response, response_ok = self.cmd("status", param_name)
        if not response_ok:
            return response, response_ok

        try:
            name, val = response.split()
            if name != param_name:
                msg = "unexpected response: {} vs {}".format(name, param_name)
                return msg, False

            return val, response_ok
        except Exception as err:
            return str(err), False

        return self.cmd("status", param_name)

    def setup(self, param_name, value):
        return self.cmd("setup", param_name, value)

    def load_setup(self, data):
        """
        Post a JSON setup to the server

        Parameters
        -----------
        data : dict
            setup info
        """
        try:
            obsmode = get_obsmode(data)
        except Exception as err:
            self.log.error("could not parse obsmode data: " + str(err))
            return False, "could not parse obsmode data"

        resp = self.cmd("seq stop")
        if not resp:
            self.log.error("could not stop sequencer")
            return False, "could not stop sequencer"

        time.sleep(0.1)
        resp = self.cmd(obsmode.readmode_command)
        if not resp:
            self.log.error("could not set read mode: " + resp)
            return False, "could not set read mode: " + resp

        time.sleep(0.1)
        for header_command in obsmode.header_commands:
            resp = self.cmd(header_command)
        if not resp:
            msg = "could not set header param {}: {} ".format(header_command, resp)
            self.log.error(msg)
            return False, msg

        time.sleep(0.1)
        resp = self.cmd(obsmode.setup_command)
        if not resp:
            self.log.error("could not setup run: " + resp)
            return False, "could not setup run: " + resp

        time.sleep(0.1)
        resp = self.cmd(obsmode.acq_command)
        if not resp:
            self.log.error("could not start acquisition process: " + resp)
            return False, "could not start acquisition process: " + resp

        if isinstance(obsmode, Idle):
            """
            A run start will start the sequencer, saving data and clearing the chips.
            There is no run start when we switch to idle mode, so start the sequencer
            manually
            """
            resp = self.cmd("seq start")
            if not resp:
                self.log.error("could not start sequencer: " + resp)
                return False, "could not start sequencer: " + resp

        return True, "All OK"

    def add_hdu(self, table_data, run_number):
        """
        Handle an uploaded table, and append it to the appropriate FITS file

        Parameters
        -----------
        table_data: bytes
            a binary encoded VOTable, which we will write to a FITS HDU,
            which is appended onto the original FITS file.
        run_number: int
            the run to append the table data to
        """
        try:
            t = ascii.read(table_data, format="ecsv")
        except Exception as e:
            self.log.error("cannot decode table data: " + str(e))
            return

        filename = os.path.join("/data", "run{:04d}.fits".format(run_number))
        if not os.path.exists(filename):
            self.log.error("no such filename exists: " + filename)
            return

        try:
            new_hdu = fits.table_to_hdu(t)
            existing_hdu_list = fits.open(filename, mode="append", memmap=True)
            existing_hdu_list.append(new_hdu)
            existing_hdu_list.flush()
        except Exception as e:
            self.log.error("could not write HDU to run: " + str(e))



# model for GPS status
from __future__ import print_function, division, unicode_literals

try:
    from subprocess import getoutput
except:
    from commands import getoutput
import os
import getpass
import warnings
import re
from twisted.logger import Logger

from astropy.time import Time

path_to_examples = "/home/{}/hipercam-gps/tsync/examples".format(getpass.getuser())


def convert(val, to_type):
    return to_type(val.split(":")[1].strip())


class GPSCall(object):
    def __init__(self, function):
        """
        Class to call a GPS function.

        Parameters
        ----------
        function: string
            Name of function to call (must be one of the example programs
            in the TSYNC library). Should include arguments
        """
        self.function = os.path.join(path_to_examples, function)

    def _check(self, response):
        """
        Implement this function to parse results of running function.

        response: string
            response from function call
        """
        if "Could not open" in response:
            raise IOError("could not open GPS device")
        if "Error" in response:
            raise IOError("error in function call:\n" + str(response))
        if "error closing" in response:
            warnings.warn("could not properly close device")

    def __call__(self):
        response = getoutput(self.function)
        self._check(response)
        return self._parse(response)


class GetNumSats(GPSCall):
    def __init__(self):
        super(GetNumSats, self).__init__("GetSatInfo 0")

    def _parse(self, response):
        response = response.strip()
        pattern = "Number of satellites tracked:\s+(\d*)"
        match = re.search(pattern, response)
        if match is None:
            raise IOError("could not parse GPS response: " + response)
        nsats = int(match.group(1))
        return nsats


class GetValidity(GPSCall):
    """
    Validity of GPS signal on time and PPS output
    """

    def __init__(self):
        super(GetValidity, self).__init__("GR_GetValidity 0 0")

    def _parse(self, response):
        time_valid, pps_valid = (
            bool(convert(val, int)) for val in response.splitlines()[2:4]
        )
        return time_valid, pps_valid


class HWTime(GPSCall):
    def __init__(self):
        super(HWTime, self).__init__("HW_GetTime 0")

    def _parse(self, response):
        year, doy, hr, mins, sec, nsec = (
            convert(val, int) for val in response.splitlines()[2:8]
        )
        time_string = "{}:{}:{}:{}:{:.6f}".format(year, doy, hr, mins, sec + nsec / 1e9)
        timestamp = Time(time_string, format="yday")
        synced = response.splitlines()[-1].split(":")[1].strip()
        # Convert to boolean
        synced = True if synced.upper() == "TRUE" else False
        return timestamp, synced


class GPSDevice(object):
    def __init__(self):
        self.get_num_sats = GetNumSats()
        self.get_validity = GetValidity()
        self.get_hw_time = HWTime()

    def telemetry(self):
        tel = dict()
        tel["num_sats"] = self.get_num_sats()
        tel["time_valid"], tel["pps_valid"] = self.get_validity()
        ts, synced = self.get_hw_time()
        ts.format = "iso"
        tel["timestamp"] = ts
        tel["synced"] = synced
        tel["state"] = {}
        return tel


class FakeGPSDevice(object):
    def telemetry(self):
        ts = Time.now()
        ts.format = "iso"
        return dict(
            num_sats=10,
            time_valid=True,
            pps_valid=True,
            timestamp=ts,
            state={},
            synced=True,
        )


class GPS(object):
    extra_rpc_calls = ()
    log = Logger()
    settable_attributes = ()

    def __init__(self, emulate=True):
        self.telemetry_delay = 10
        self.emulate = emulate
        # no state machine needed
        self.machines = {}
        if self.emulate:
            self.device = FakeGPSDevice()
        else:
            self.device = GPSDevice()

    def telemetry(self):
        return self.device.telemetry()



from functools import partial

from astropy import units as u
from astropy.time import Time
from twisted.logger import Logger

from ..devices.meerstetter import (
    STAGE_ENABLE_CODES,
    STATUS_CODES,
    MeerstetterTEC1090,
    MeerstetterTEC1090Emulator,
)
from ..machines import create_machine

TEMP_UNITS = u.deg_C
CURRENT_UNITS = u.A


# define some helper functions to create sensor statemachines
def create_sensor(connection_handler, address, name, read_method, initial_value):
    sensor = create_machine(
        "sensor",
        initial_context={
            "value": initial_value,
            "read_value": read_method,
            "name": name,
        },
    )
    connection_handler.bind(sensor)
    sensor.bind(connection_handler)
    return sensor


def create_controller(
    connection_handler, address, name, read_method, write_method, setpoint_read_method
):
    sensor = create_machine(
        "controller",
        initial_context={
            "write": write_method,
            "read": read_method,
            "name": name,
            "read_setpoint": setpoint_read_method,
        },
    )
    connection_handler.bind(sensor)
    sensor.bind(connection_handler)
    return sensor


class Meerstetter(object):
    """
    Meerstetter peltier controller.

    This is a model for a single peltier controller, that has three channels.
    """

    log = Logger()

    # can be set by publishing to device level topic of same name
    settable_attributes = (
        "target_temperature1",
        "target_temperature2",
        "target_temperature3",
        "stage_enable1",
        "stage_enable2",
        "stage_enable3",
    )

    # force the stage_enable mode to be "HW Enable" for all channels
    # so that peltier stops when flow disconnects
    extra_looping_calls = {"force_hw_enable": 60}

    # all state change triggers are automatically addded as RPC calls
    # add any extra methods you want exposed here
    extra_rpc_calls = ("reset",)

    def __init__(self, ip_addr, port, emulate=True):

        if emulate:
            dev = MeerstetterTEC1090Emulator(ip_addr, port)
            self.dev = dev
        else:
            dev = MeerstetterTEC1090(ip_addr, port)
            self.dev = dev

        self._controller = create_machine(
            "connections", initial_context={"device": self, "name": "meerstetter"}
        )
        self.machines = {"connection": self._controller}

        for addr in range(1, 4):
            name = "status{}".format(addr)
            self.machines[name] = create_sensor(
                self._controller, addr, name, partial(self.dev.get_status, addr), 0.0
            )

            name = "heatsink{}".format(addr)
            self.machines[name] = create_sensor(
                self._controller,
                addr,
                name,
                partial(self.dev.get_heatsink_temp, addr),
                0.0,
            )

            name = "current{}".format(addr)
            self.machines[name] = create_sensor(
                self._controller, addr, name, partial(self.dev.get_current, addr), 0.0
            )

            name = "temperature{}".format(addr)
            self.machines[name] = create_controller(
                self._controller,
                addr,
                name,
                partial(self.dev.get_ccd_temp, addr),
                partial(self.dev.set_ccd_temp, addr),
                partial(self.dev.get_setpoint, addr),
            )

            name = "stage_enable{}".format(addr)
            self.machines[name] = create_controller(
                self._controller,
                addr,
                name,
                partial(self.dev.get_stage_enable, addr),
                partial(self.dev.set_stage_enable, addr),
                partial(self.dev.get_stage_enable, addr),
            )

    @property
    def target_temperature1(self):
        return self.machines["temperature1"].context["target"]

    @target_temperature1.setter
    def target_temperature1(self, value):
        """
        Set target. Assume raw floats are already in degrees
        """
        if isinstance(value, u.Quantity):
            value = value.to_value(TEMP_UNITS, equivalencies=u.temperature())
        self.machines["temperature1"].queue("targetSet", target=value)

    @property
    def target_temperature2(self):
        return self.machines["temperature2"].context["target"]

    @target_temperature2.setter
    def target_temperature2(self, value):
        if isinstance(value, u.Quantity):
            value = value.to_value(TEMP_UNITS, equivalencies=u.temperature())
        self.machines["temperature2"].queue("targetSet", target=value)

    @property
    def target_temperature3(self):
        return self.machines["temperature3"].context["target"]

    @target_temperature3.setter
    def target_temperature3(self, value):
        if isinstance(value, u.Quantity):
            value = value.to_value(TEMP_UNITS, equivalencies=u.temperature())
        self.machines["temperature3"].queue("targetSet", target=value)

    @property
    def stage_enable1(self):
        return self.machines["stage_enable1"].context["value"]

    @stage_enable1.setter
    def stage_enable1(self, value):
        value = int(value)
        self.machines["stage_enable1"].queue("targetSet", target=value)

    @property
    def stage_enable2(self):
        return self.machines["stage_enable2"].context["value"]

    @stage_enable2.setter
    def stage_enable2(self, value):
        value = int(value)
        self.machines["stage_enable2"].queue("targetSet", target=value)

    @property
    def stage_enable3(self):
        return self.machines["stage_enable3"].context["value"]

    @stage_enable3.setter
    def stage_enable3(self, value):
        value = int(value)
        self.machines["stage_enable3"].queue("targetSet", target=value)

    def reset(self, address):
        self.dev.reset_tec(address)

    def _oldest_timestamp(self):
        """
        search all the sensors and return the oldest (stalest) timestamp
        """
        stamps = []
        for key in self.machines:
            if "last_read" in self.machines[key].context:
                stamps.append(self.machines[key].context["last_read"])
        return min(stamps)

    def telemetry(self):
        ts = Time(self._oldest_timestamp(), format="unix")
        ts.format = "iso"
        tel = dict(
            timestamp=ts,
            state={key: self.machines[key].configuration for key in self.machines},
        )
        for key in self.machines:
            if key.startswith("status"):
                tel[key] = STATUS_CODES[self.machines[key].context["value"]]
            elif key.startswith("stage_enable"):
                tel[key] = STAGE_ENABLE_CODES[self.machines[key].context["value"]]
            elif key != "connection":
                unit = CURRENT_UNITS if key.startswith("current") else TEMP_UNITS
                tel[key] = self.machines[key].context["value"] * unit
        for addr in range(1, 4):
            key = "target_temperature{}".format(addr)
            tel[key] = getattr(self, key) * TEMP_UNITS
        return tel

    def force_hw_enable(self):
        """
        Set all outputs to HW Enable so that peltier stops when flow disconnects
        """
        try:
            self.stage_enable1 = 3
            self.stage_enable2 = 3
            self.stage_enable3 = 3
        except Exception as e:
            self.log.error("failed to force hw enable: {error}", error=e)

    # METHODS NEEDED FOR STATE MACHINES
    def try_connect(self):
        """
        Always needs to be defined.  Returns True on successful connection attempt.
        """
        # attempt connection, return true to allow device state to change
        try:
            connected = self.dev.connect()
        except Exception:
            return False
        return connected

    def disconnect(self):
        self.dev.disconnect()
